from findpeaks import findpeaks
import numpy as np
import unittest

class TestFINDPEAKS(unittest.TestCase):

    def test_fit(self):
        # CHECK OUTPUT METHOD TOPOLOGY
        import numpy as np
        import matplotlib.pyplot as plt
        from findpeaks import findpeaks
        fp = findpeaks(method="topology")
        X = fp.import_example('2dpeaks')
        results = fp.fit(X)
        assert fp.type=='peaks2d'
        assert [*results.keys()]==['Xraw', 'Xproc', 'Xdetect', 'Xranked', 'persistence', 'groups0']
        assert [*fp.args]==['limit', 'scale', 'denoise', 'togray', 'imsize', 'figsize', 'type']
        assert results['Xraw'].shape==results['Xdetect'].shape
        assert results['Xproc'].shape==results['Xdetect'].shape
    
        # CHECK RESULTS METHOD TOPOLOGY
        # assert len(results['peak'])==20
        assert len(results['Xdetect'][results['Xdetect']!=0])==18
        assert len(results['Xranked'][results['Xranked']!=0])==21
        assert np.sum(results['Xdetect'][results['Xranked']!=0]>0)==18
    
        # CHECK RESULTS METHOD with LIMIT functionality
        fp = findpeaks(method="topology", limit=0)
        X = fp.import_example('2dpeaks')
        results = fp.fit(X)
        assert len(results['Xdetect'][results['Xdetect']!=0])==18
        assert len(results['Xranked'][results['Xranked']!=0])==18
        
        # CHECK OUTPUT METHOD MASK
        fp = findpeaks(method="mask", verbose=3)
        X = fp.import_example('2dpeaks')
        results = fp.fit(X)
        assert fp.type=='peaks2d'
        assert [*results.keys()]==['Xraw', 'Xproc', 'Xdetect','Xranked']
        assert [*fp.args]==['limit', 'scale', 'denoise', 'togray', 'imsize', 'figsize', 'type']
    
        # CHECK RESULTS METHOD TOPOLOGY
        assert np.sum(results['Xdetect'])==20
        assert results['Xraw'].shape==results['Xdetect'].shape
        assert results['Xproc'].shape==results['Xdetect'].shape
    
        # CHECK WHITELIST
        import numpy as np
        from scipy.ndimage import gaussian_filter
        from findpeaks import findpeaks
        rng = np.random.default_rng(42)
        x = rng.normal(size=(50, 50))
        x = gaussian_filter(x, sigma=10.)
        # peak and valley
        fp = findpeaks(method="topology", whitelist=['peak', 'valley'], denoise=None, verbose=3)
        results = fp.fit(x)
        assert results['persistence']['peak'].sum()==4
        assert results['persistence']['valley'].sum()==4
        assert np.sum(results['Xdetect']>0)==6
        assert np.sum(results['Xdetect']<0)==0
        assert np.sum(results['Xranked']>0)==4
        assert np.sum(results['Xranked']<0)==4
    
        # valley
        fp = findpeaks(method="topology", whitelist='peak', denoise=None, verbose=3)
        results = fp.fit(x)
        assert results['persistence']['peak'].shape[0]==results['persistence']['peak'].sum()
        assert np.sum(results['Xdetect']>0)==3
        assert np.sum(results['Xdetect']<0)==0
        assert np.sum(results['Xranked']>0)==4
        assert np.sum(results['Xranked']<0)==0
    
        fp = findpeaks(method="topology", whitelist='valley', denoise=None, verbose=3)
        results = fp.fit(x)
        assert results['persistence']['valley'].shape[0]==results['persistence']['valley'].sum()
        assert np.sum(results['Xdetect']>0)==3
        assert np.sum(results['Xdetect']<0)==0
        assert np.sum(results['Xranked']>0)==0
        assert np.sum(results['Xranked']<0)==4
    
        # CHECK OUTPUT METHOD TOPOLOGY
        fp = findpeaks(method="topology")
        X = fp.import_example('1dpeaks')
        results = fp.fit(X)
        assert fp.type=='peaks1d'
        assert [*results.keys()]==['persistence', 'Xdetect', 'Xranked', 'groups0', 'df']
        assert [*fp.args]==['method', 'params', 'lookahead', 'interpolate', 'figsize', 'type']
        assert len(X)==len(results['Xdetect'])
        assert len(X)==len(results['Xranked'])
        assert len(X)==results['df'].shape[0]
        assert np.all(np.isin(results['df'].columns, ['x', 'y', 'labx', 'rank', 'score', 'valley', 'peak']))
        assert np.all(np.isin(results['persistence'].columns, ['x', 'y', 'birth_level', 'death_level', 'score']))
        
        # CHECK RESULTS METHOD TOPOLOGY
        assert results['persistence'].shape[0]==7
        assert len(results['Xdetect'][results['Xdetect']!=0])==7
        assert len(results['Xranked'][results['Xranked']!=0])==7
        assert np.sum(results['Xdetect'][results['Xranked']!=0]>0)==7
    
        # CHECK RESULTS METHOD with LIMIT functionality
        fp = findpeaks(method="topology", limit=0.02)
        results = fp.fit(X)
        assert results['persistence'].shape[0]==4
        assert len(results['Xdetect'][results['Xdetect']!=0])==len(results['Xranked'][results['Xranked']!=0])
        assert np.sum(results['Xdetect'][results['Xranked']!=0]>0)==4
    
        
        # CHECK OUTPUT METHOD PEAKDETECT
        fp = findpeaks(method="peakdetect", lookahead=1, verbose=3)
        X = fp.import_example('1dpeaks')
        results = fp.fit(X)
        assert fp.type=='peaks1d'
        assert [*results.keys()]==['df']
        assert [*fp.args]==['method', 'params', 'lookahead', 'interpolate', 'figsize', 'type']
        assert len(X)==results['df'].shape[0]
        assert np.all(np.isin(results['df'].columns, ['x', 'y', 'labx', 'valley', 'peak', 'rank', 'score']))
        
        # CHECK RESULTS METHOD TOPOLOGY
        assert results['df']['peak'].sum()==2
        assert results['df']['valley'].sum()==4
    
        # Run over all combinations and make sure no errors are made
        X = [10,11,9,23,21,11,45,20,11,12]
        methods = ['topology', 'peakdetect', None]
        interpolates = [None, 1, 10, 1000]
        lookaheads =[None, 0, 1, 10, 100]
        for method in methods:
            for interpolate in interpolates:
                for lookahead in lookaheads:
                    fp = findpeaks(lookahead=lookahead, interpolate=interpolate, method=method, verbose=0)
                    assert fp.fit(X)
    
    
    def test_denoising(self):

        # DENOISING METHODS TEST
        from findpeaks import findpeaks
        fp = findpeaks()
        img = fp.import_example('2dpeaks_image')
        import findpeaks
        
        # filters parameters
        # window size
        winsize = 15
        # damping factor for frost
        k_value1 = 2.0
        # damping factor for lee enhanced
        k_value2 = 1.0
        # coefficient of variation of noise
        cu_value = 0.25
        # coefficient of variation for lee enhanced of noise
        cu_lee_enhanced = 0.523
        # max coefficient of variation for lee enhanced
        cmax_value = 1.73
        
        # Some pre-processing
        # Resize
        img = findpeaks.stats.resize(img, size=(300,300))
        # Make grey image
        img = findpeaks.stats.togray(img)
        # Scale between [0-255]
        img = findpeaks.stats.scale(img)
        
        # Denoising
        # fastnl
        img_fastnl = findpeaks.stats.denoise(img.copy(), method='fastnl', window=winsize)
        # bilateral
        img_bilateral = findpeaks.stats.denoise(img.copy(), method='bilateral', window=winsize)
        # frost filter
        image_frost = findpeaks.stats.frost_filter(img.copy(), damping_factor=k_value1, win_size=winsize)
        # kuan filter
        image_kuan = findpeaks.kuan_filter(img.copy(), win_size=winsize, cu=cu_value)
        # lee filter
        image_lee = findpeaks.stats.lee_filter(img.copy(), win_size=winsize, cu=cu_value)
        # lee enhanced filter
        image_lee_enhanced = findpeaks.stats.lee_enhanced_filter(img.copy(), win_size=winsize, k=k_value2, cu=cu_lee_enhanced, cmax=cmax_value)
        # lee sigma filter
        image_lee_sigma = findpeaks.stats.lee_sigma_filter(img.copy(), win_size=winsize)
        # mean filter
        image_mean = findpeaks.stats.mean_filter(img.copy(), win_size=winsize)
        # median filter
        image_median = findpeaks.stats.median_filter(img.copy(), win_size=winsize)
    
        
        # Loop throughout many combinations of parameter settings
        from findpeaks import findpeaks
        methods = ['caerus', 'mask','topology', None]
        filters = ['lee','lee_enhanced','lee_sigma','kuan','fastnl','bilateral','frost','median','mean', None]
        windows = [None, 3, 63]
        cus = [None, 0, 0.75]
        img = fp.import_example('2dpeaks')
    
        for getfilter in filters:
            for window in windows:
                for cu in cus:
                    fp = findpeaks(method='topology', scale=True, denoise=getfilter, params={'window': window, 'cu': cu}, togray=True, imsize=None, verbose=3)
                    assert fp.fit(img)
                    # assert fp.plot_mesh(wireframe=False)
                    # plt.close('all')
                    # assert fp.plot_persistence()
                    # plt.close('all')
                    # assert fp.plot()
                    # plt.close('all')
                    # assert fp.plot_preprocessing()



from findpeaks.findpeaks import findpeaks

from findpeaks.findpeaks import (
    import_example,
    stats,
    interpolate,
    )

# Import the denosing filters
from findpeaks.filters.lee import lee_filter
from findpeaks.filters.lee_enhanced import lee_enhanced_filter
from findpeaks.filters.lee_sigma import lee_sigma_filter
from findpeaks.filters.kuan import kuan_filter
from findpeaks.filters.frost import frost_filter
from findpeaks.filters.median import median_filter
from findpeaks.filters.mean import mean_filter

__author__ = 'Erdogan Tasksen'
__email__ = 'erdogant@gmail.com'
__version__ = '2.5.3'

# module level doc-string
__doc__ = """
findpeaks
=====================================================================

findpeaks is for the detection and vizualization of peaks and valleys in a 1D-vector and 2D-array.
In case of 2D-array, the image can be pre-processed by resizing, scaling, and denoising.
Peaks are detected using the masking and the topology method.
The results can be plotted for the preprocessing steps, the persistence of peaks, the final masking plot and a 3d-mesh plot.

Examples
--------
>>> from findpeaks import findpeaks
>>> X = [9,60,377,985,1153,672,501,1068,1110,574,135,23,3,47,252,812,1182,741,263,33]
>>> fp = findpeaks(interpolate=10, lookahead=1)
>>> results = fp.fit(X)
>>> fp.plot()
>>>
>>> # 2D array example
>>> from findpeaks import findpeaks
>>> fp = findpeaks(method='topology')
>>> X = fp.import_example('2dpeaks')
>>> results = fp.fit(X)
>>> fp.plot()
>>>
>>> # Image example
>>> from findpeaks import findpeaks
>>> fp = findpeaks(method='topology', denoise='fastnl', window=30, imsize=(300, 300))
>>> X = fp.import_example('2dpeaks_image')
>>> results = fp.fit(X)
>>> fp.plot()
>>>
>>> # Plot each seperately
>>> fp.plot_preprocessing()
>>> fp.plot_persistence()
>>> fp.plot_mesh()

References
----------
    * https://github.com/erdogant/findpeaks
    * https://www.sthu.org/code/codesnippets/imagepers.html
    * H. Edelsbrunner and J. Harer, Computational Topology. An Introduction, 2010, ISBN 0-8218-4925-5.
    * https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.find_peaks.html

"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 2023: Caroline Goehner: <carolinesophie.goehner@eurac.edu> <carosophie.goehner@gmail.com>
# This part of the library was written at Eurac Research in the 
# framework of the project ScaleAgData (SCALING ΑGRICULTURAL SENSOR 
# DATA for an improved monitoring of agri-environmental conditions. 
# Duration: 01/01/2023 - 31/12/2026) funded by the Horizon Europe 
# program under the grant agreement no 101086355.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import xarray as xr
from joblib import Parallel, delayed


sigma_DEFAULT = 0.9 # for general applications
win_size_DEFAULT = 7
num_looks_DEFAULT = 1
tk_DEFAULT = 5 # as in S1TBX
num_cores_DEFAULT = -1 
data_measure_DEFAULT = "intensity"

def assert_parameters(sigma, win_size, num_looks, tk, data_measure):
    """
    Asserts parameters in range.
    Parameters:
        - sigma: in [0.5, 0.6, 0.7, 0.8, 0.9]
        - win_size: should be odd, at least 3
        - num_looks: in [1, 2, 3, 4]
        - tk: in [5, 6, 7]
        - data_measure: "intensity" or "amplitude"
    """

    if sigma not in [0.5, 0.6, 0.7, 0.8, 0.9]: raise Exception("Sigma parameter has to be 0.5, 0.6, 0.7, 0.8, or 0.9, submitted %s" %(sigma))
    if win_size < 3: raise Exception('ERROR: win size must be at least 3')
    if num_looks not in [1, 2, 3, 4]: raise Exception("num_looks parameter has to be 1, 2, 3 or 4, submitted %s" %(num_looks))
    if tk not in [5, 6, 7]: print('[findpeaks] >For general applications it is recommended to use threshold tk between 5 and 7. You provided %s.' %(tk))
    if data_measure not in ["intensity", "amplitude"]: raise Exception('ERROR: data_measure has to be "intensity" or "amplitude". You provided %s.' %(data_measure))


def ptTar(x, y, img, Z98, tk):
    """
    Detect if the pixel is part of a point target of surrounding pixels
    Parameters:
        - x: int
            X-coordinate of the pixel.
        - y: int
            Y-coordinate of the pixel.
        - img: xarray
            Input image.
        - Z98: ndarray
            Threshold of the 98th percentile of the img.
        - tk: int
            Threshold for number of K neighbouring pixels > Z98 to classify the pixel as point target, typically 5.
    """

    for c in [[-1, -1], [-1, 0], [-1, 1], [0, -1], [0, 1], [1, -1], [1, 0], [1, 1]]:
        a = x+c[0]
        b = y+c[1]
        win = img[a-1:a+1, b-1:b+1] # 3x3 windows for pixels surrounding the center pixel
        K_win = np.count_nonzero(win >= Z98) # number of pixels outside the Z98
        if K_win >= tk: # is point target
            ptTarget = True
            break
        else: 
            ptTarget = False
            continue
    return(ptTarget)


def lee_sigma_filter(img,
                     sigma = sigma_DEFAULT,
                     win_size = win_size_DEFAULT,
                     num_looks = num_looks_DEFAULT,
                     tk = tk_DEFAULT,
                     num_cores = num_cores_DEFAULT,
                     data_measure = data_measure_DEFAULT): 
    """Lee sigma filter.

    Description
    -----------
    Improved Lee Sigma, according to Lee Sigma filter in SNAP Sentinel-1 Toolbox.
    Apply the filter with a window of win_size x win_size to a numpy matrix (containing the image), before converting to dB.
    Jong-Sen Lee, Jen-Hung Wen, T. L. Ainsworth, Kun-Shan Chen and A. J. Chen, "Improved Sigma Filter for Speckle Filtering of SAR Imagery",
    in IEEE Transactions on Geoscience and Remote Sensing, vol. 47, no. 1, pp. 202-213, Jan. 2009, doi: 10.1109/TGRS.2008.2002881.

    Parameters
    ----------
    img : numpy.ndarray or xarray.DataArray
        Input image.
    sigma : float, (default: 0.9)
        Speckle noise standard deviation.
    win_size : int, int (default: 7)
        Window size.
    num_looks : int, (default: 1)
        Number of looks of the SAR img.
    tk: int, (default: 5)
        Threshold of neighbouring pixels outside of the 98th percentile, typically between 5 and 7.
    num_cores: int, (default: -1)
        Number of cores to use for parallel computing, if -1 all CPUs are used, if 1 no parallel computing is used.

    Returns
    -------
    img_filtered : numpy.ndarray or xarray.DataArray
        Filtered image, type depending on input type.

    Examples
    --------
    >>> import findpeaks
    >>> import matplotlib.pyplot as plt
    >>> img = findpeaks.import_example('2dpeaks_image')
    >>> # Resize
    >>> img = findpeaks.stats.resize(img, size=(300,300))
    >>> # Make grey image
    >>> img = findpeaks.stats.togray(img)
    >>> # Scale between [0-255]
    >>> img = findpeaks.stats.scale(img)
    >>> # Filter
    >>> img_filtered = findpeaks.stats.lee_sigma_filter(img.copy(), win_size=7)
    >>>
    >>> plt.figure()
    >>> fig, axs = plt.subplots(1,2)
    >>> axs[0].imshow(img, cmap='gray'); axs[0].set_title('Input')
    >>> axs[1].imshow(img_filtered, cmap='gray'); axs[1].set_title('Lee sigma filter')

    """

    if win_size < 3: raise Exception('[findpeaks] >ERROR: win size must be at least 3')
    if len(img.shape) > 2: raise Exception('[findpeaks] >ERROR: Image should be 2D. Hint: set the parameter: togray=True')
    if ((win_size % 2) == 0): print('[findpeaks] >It is highly recommended to use odd window sizes. You provided %s, an even number.' % (win_size))
    assert_parameters(sigma, win_size, num_looks, tk, data_measure) # check validity of input parameters

    if data_measure == "intensity":
        if num_looks == 1:
            if sigma == 0.5:
                I1 = 0.436 # lower sigma range
                I2 = 1.920 # upper sigma range
                IsigmaVP = 0.4057 # speckle noise standard deviation (adjusted)

            elif sigma == 0.6:
                I1 = 0.343
                I2 = 2.210
                IsigmaVP = 0.4954
            elif sigma == 0.7:
                I1 = 0.254
                I2 = 2.582
                IsigmaVP = 0.5911
            elif sigma == 0.8:
                I1 = 0.168
                I2 = 3.094
                IsigmaVP = 0.6966
            elif sigma == 0.9:
                I1 = 0.084
                I2 = 3.941
                IsigmaVP = 0.8191

        elif num_looks == 2:
            if sigma == 0.5:
                I1 = 0.582
                I2 = 1.584
                IsigmaVP = 0.2763
            elif sigma == 0.6:
                I1 = 0.501
                I2 = 1.755
                IsigmaVP = 0.3388
            elif sigma == 0.7:
                I1 = 0.418
                I2 = 1.972
                IsigmaVP = 0.4062
            elif sigma == 0.8:
                I1 = 0.327
                I2 = 2.260
                IsigmaVP = 0.4810
            elif sigma == 0.9:
                I1 = 0.221
                I2 = 2.744
                IsigmaVP = 0.5699

        elif num_looks == 3:
            if sigma == 0.5:
                I1 = 0.652
                I2 = 1.458
                IsigmaVP = 0.2222
            elif sigma == 0.6:
                I1 = 0.580
                I2 = 1.586
                IsigmaVP = 0.2736
            elif sigma == 0.7:
                I1 = 0.505
                I2 = 1.751
                IsigmaVP = 0.3280
            elif sigma == 0.8:
                I1 = 0.419
                I2 = 1.965
                IsigmaVP = 0.3892
            elif sigma == 0.9:
                I1 = 0.313
                I2 = 2.320
                IsigmaVP = 0.4624

        elif num_looks == 4:
            if sigma == 0.5:
                I1 = 0.694
                I2 = 1.385
                IsigmaVP = 0.1921
            elif sigma == 0.6:
                I1 = 0.630
                I2 = 1.495
                IsigmaVP = 0.2348
            elif sigma == 0.7:
                I1 = 0.560
                I2 = 1.627
                IsigmaVP = 0.2825
            elif sigma == 0.8:
                I1 = 0.480
                I2 = 1.804
                IsigmaVP = 0.3354
            elif sigma == 0.9:
                I1 = 0.378
                I2 = 2.094
                IsigmaVP = 0.3991
    
    elif data_measure == "amplitude":
        if num_looks == 1:
            if sigma == 0.5:
                A1 = 0.653997
                A2 = 1.40002
                AsigmaVP = 0.208349
            elif sigma == 0.6:
                A1 = 0.578998
                A2 = 1.50601
                AsigmaVP = 0.255358
            elif sigma == 0.7:
                A1 = 0.496999
                A2 = 1.63201
                AsigmaVP = 0.305303
            elif sigma == 0.8:
                A1 = 0.403999
                A2 = 1.79501
                AsigmaVP = 0.361078
            elif sigma == 0.9:
                A1 = 0.286
                A2 = 2.04301
                AsigmaVP = 0.426375
        
        elif num_looks == 2:
            if sigma == 0.5:
                A1 = 0.76
                A2 = 1.263
                AsigmaVP = 0.139021
            elif sigma == 0.6:
                A1 = 0.705
                A2 = 1.332
                AsigmaVP = 0.169777
            elif sigma == 0.7:
                A1 = 0.643
                A2 = 1.412
                AsigmaVP = 0.206675
            elif sigma == 0.8:
                A1 = 0.568
                A2 = 1.515
                AsigmaVP = 0.244576
            elif sigma == 0.9:
                A1 = 0.467
                A2 = 1.673
                AsigmaVP = 0.29107
        
        elif num_looks == 3:
            if sigma == 0.5:
                A1 = 0.806
                A2 = 1.21
                AsigmaVP = 0.109732
            elif sigma == 0.6:
                A1 = 0.76
                A2 = 1.263
                AsigmaVP = 0.138001
            elif sigma == 0.7:
                A1 = 0.708
                A2 = 1.327
                AsigmaVP = 0.163686
            elif sigma == 0.8:
                A1 = 0.645
                A2 = 1.408
                AsigmaVP = 0.19597
            elif sigma == 0.9:
                A1 = 0.557
                A2 = 1.531
                AsigmaVP = 0.234219
        
        elif num_looks == 4:
            if sigma == 0.5:
                A1 = 0.832
                A2 = 1.179
                AsigmaVP = 0.0894192
            elif sigma == 0.6:
                A1 = 0.793
                A2 = 1.226
                AsigmaVP = 0.112018
            elif sigma == 0.7:
                A1 = 0.747
                A2 = 1.279
                AsigmaVP = 0.139243
            elif sigma == 0.8:
                A1 = 0.691
                A2 = 1.347
                AsigmaVP = 0.167771
            elif sigma == 0.9:
                A1 = 0.613
                A2 = 1.452
                AsigmaVP = 0.201036

    # variables
    final_img = None
    if isinstance(img, xr.DataArray): # make it possible to use xarray dataarrays as well 
        final_img = img.copy()
        img = img.values
    win_size_h = int(win_size/2) # "half" window as distance from center pixel in each direction
    
    if data_measure == "intensity":
        sigmaV = 1.0 / (num_looks ** 0.5) # standard deviation of the multiplicative speckle noise, depending on number of looks
        sigmaVP = IsigmaVP
        sigmaRangeLow = I1
        sigmaRangeHigh = I2
    elif data_measure == "amplitude":
        sigmaV = 0.5227 / (num_looks ** 0.5)
        sigmaVP = AsigmaVP
        sigmaRangeLow = A1
        sigmaRangeHigh = A2
    sigmaVSqr = sigmaV**2 # variance of the multiplicative speckle noise
    
    Z98 = np.percentile(img, 98) # threshold of the 98th percentile of the SAR img
    N, M = img.shape
    img_filtered = np.zeros_like(img, dtype=float)  

    def filter_pixel(i, j):
        xleft = i - win_size_h # define left x coordinate of the selected window size
        xright = i + win_size_h+1 # define right x coordinate of the selected window size, add 1 for indexing ndarrays
        if xleft < 0: xleft = 0 # if outside the image dimensions set to min x coordinate
        if xright >= N: xright = N # if outside the image dimensions set to max x coordinate
        
        xleft3 = i - 1 # for 3x3 window
        xright3 = i + 2
        if xleft3 < 0: xleft3 = 0
        if xright3 >= N: xright3 = N
        
        yup = j - win_size_h # in y dimension
        ydown = j + win_size_h+1 
        if yup < 0: yup = 0
        if ydown >= M: ydown = M

        yup3 = j - 1 # for 3x3 window
        ydown3 = j + 2
        if yup3 < 0: yup3 = 0
        if ydown3 >= M: ydown3 = M

        # 1. Point target detection + preservation 
        z = img[i, j] # center pixel value of window
        window = img[xleft:xright, yup:ydown] # window of selected size
        window_3x3 = img[xleft3:xright3, yup3:ydown3]  # 3x3 window

        K = np.count_nonzero(window_3x3 >= Z98) # number of pixels in the 3x3 window outside the Z98

        if (ptTar(i, j, img, Z98, tk) == False # not part of a (earlier) point target
            and (z.item() >= Z98) == False # pixel value is within the 98th percentile of the SAR img
            or ((z.item() >= Z98) == True and (K >= tk) == False) # is not in the 98th percentile, but has enough surrounding pixels that are neither -> it will be filtered
           ): 

            # 2. Pixels selection based on the sigma range
            # - MMSE on 3x3 using orig_sigmaVP to compute a priori mean (priori_x)                   
            mean_z = window_3x3.mean() # local mean in 3x3
            Var_z = window_3x3.var(dtype = np.float64) # local variance in 3x3
            Var_x = (Var_z - mean_z**2 * sigmaVSqr) / (1 + sigmaVSqr) # Variance of x
            if Var_x < 0: Var_x = 0.0 # according to s1tbx
            b = Var_x / (Var_z+1e-50) # weight function - add small values to avoid nan weights when all the values are similar in the window

            priori_x = (1-b) * mean_z + b * z  # MMSE filter to calculate a priori mean

            # - establish sigma range using LUT for sigma in Intensity img and num_looks:
            XsigmaRangeLow = sigmaRangeLow * priori_x # lower sigma range
            XsigmaRangeHigh = sigmaRangeHigh * priori_x # upper sigma range
            sigmaVPSqr = sigmaVP**2 # speckle noise variance

            # - select pixels in window if their values fall into sigma range, compute mean_z and Var_z
            window = window[np.where(np.logical_and(window >= XsigmaRangeLow, window <= XsigmaRangeHigh))]
            if np.count_nonzero(window) == 0: new_pix_value = z # when window is empty, according to S1TBX
            else:
                mean_z = window.mean() # local mean in the sigma range
                Var_z = window.var(dtype = np.float64) # local variance in the sigma range

                # 3. MMSE application
                # - compute MMSE filter weight b using Var_x, based on mean_z, Var_z and sigmaVPSqr
                Var_x = (Var_z - mean_z**2 * sigmaVPSqr) / (1 + sigmaVPSqr) # Variance of x
                if Var_x < 0: Var_x = 0.0 # according to s1tbx
                b = Var_x / (Var_z+1e-50) # weight function - add small values to avoid nan weights when all the values are similar in the window

                # - filter center pixel using MMSE
                new_pix_value = (1-b) * mean_z + b * z # new filtered pixel value
            

        else: # center pixel is part of a (earlier) point target or is a point target pixel -> it will NOT be filtered
            new_pix_value = z
            
        return new_pix_value
    
    # Parallel Process
    result = Parallel(n_jobs=num_cores)(
        delayed(filter_pixel)(i, j) for i in range(N) for j in range(M)
    )

    # Unpack the results 
    for (index, v), value in zip(np.ndenumerate(img_filtered), result):
        img_filtered[index[0], index[1]] = value

    if isinstance(final_img, xr.DataArray): # in case xarray dataarray was used
        final_img.values = img_filtered
        return final_img
    else:
        return img_filtered



"""Python library for the detection of peaks and valleys."""
# ----------------------------------------------------
# Name        : findpeaks.py
# Author      : E.Taskesen
# Contact     : erdogant@gmail.com
# github      : https://github.com/erdogant/findpeaks
# Licence     : See LICENSE
# ----------------------------------------------------

# import findpeaks
from peakdetect import peakdetect
from caerus import caerus
import caerus.utils.csplots as csplots
import matplotlib.pyplot as plt
from tqdm import tqdm
import pandas as pd
import numpy as np
import os
import requests
from urllib.parse import urlparse

# #### DEBUG ONLY ####
# import stats as stats
# from stats import disable_tqdm
# import interpolate as interpolate
# #####################
import findpeaks.stats as stats
from findpeaks.stats import disable_tqdm
import findpeaks.interpolate as interpolate
# #####################


# %%
class findpeaks():
    """Python library for the detection of peaks and valleys.

    findpeaks is for the detection and vizualization of peaks and valleys in a 1D-vector and 2D-array.
    In case of 2D-array, the image can be pre-processed by resizing, scaling, and denoising. For a 1D-vector,
    pre-processing by interpolation is possible. Peaks can be detected using various methods, and the results can be
    vizualized, such as the preprocessing steps, the persistence of peaks, the masking plot and a 3d-mesh plot.

    Examples
    --------
    >>> from findpeaks import findpeaks
    >>> X = [9,60,377,985,1153,672,501,1068,1110,574,135,23,3,47,252,812,1182,741,263,33]
    >>> fp = findpeaks(method='peakdetect', interpolate=10, lookahead=1)
    >>> results = fp.fit(X)
    >>> fp.plot()
    >>>
    >>> # 2D array example
    >>> from findpeaks import findpeaks
    >>> X = fp.import_example('2dpeaks')
    >>> results = fp.fit(X)
    >>> fp.plot()
    >>>
    >>> # Image example
    >>> from findpeaks import findpeaks
    >>> fp = findpeaks(method='topology', denoise='fastnl', params={'window': 30}, imsize=(300,300))
    >>> X = fp.import_example('2dpeaks_image')
    >>> results = fp.fit(X)
    >>> fp.plot()
    >>>
    >>> # Plot each seperately
    >>> fp.plot_preprocessing()
    >>> fp.plot_persistence()
    >>> fp.plot_mesh()

    References
    ----------
        * https://erdogant.github.io/findpeaks/

    """

    def __init__(self,
                 method=None,
                 whitelist=['peak', 'valley'],
                 lookahead=200,
                 interpolate=None,
                 limit=None,
                 imsize=None,
                 scale=True,
                 togray=True,
                 denoise='fastnl',
                 window=None,  # DEPRECATED IN LATER VERSIONS: specify in params
                 cu=None,  # DEPRECATED IN LATER VERSIONS: specify in params
                 params_caerus={},  # DEPRECATED IN LATER VERSIONS: use params instead
                 params={'window': 3},
                 figsize=(15, 8),
                 verbose=3):
        """Initialize findpeaks parameters.

        Parameters
        ----------
        X : array-like (1d-vector or 2d-image)
            Input image data.
        method : String, (default : None).
            Available methods for peak detection. In case method=None, the default is choosen.
            1d-vector approaches:
                * 'topology'
                * 'peakdetect' (default)
                * 'caerus'
            2d-array approaches:
                * 'topology' (default)
                * 'mask'
        whitelist : str or list ['peak','valley']
            Choose what to detect:
                * 'peak'
                * 'valley'
                * ['peak','valley']
        lookahead : int, (default : 200)
            Looking ahead for peaks. For very small 1d arrays (such as up to 50 datapoints), use low numbers such as 1 or 2.
        interpolate : int, (default : None)
            Interpolation factor. The higher the number, the less sharp the edges will be.
        limit : float, (default : None)
            In case method='topology'
            Values > limit are active search areas to detect regions of interest (ROI).
        imsize : tuple, (default : None)
            resize to (width,length).
        scale : bool, (default : False)
            Scaling in range [0-255] by img*(255/max(img))
        denoise : string, (default : 'fastnl', None to disable)
            Filtering method to remove noise:
                * None
                * 'fastnl'
                * 'bilateral'
                * 'lee'
                * 'lee_enhanced'
                * 'lee_sigma'
                * 'kuan'
                * 'frost'
                * 'median'
                * 'mean'
        params : dict():
            Denoising parameters for the methods. If None are defined, the default will be used:
            caerus (default): {'window': 50, 'minperc': 3, 'nlargest': 10, 'threshold': 0.25}
            lee_sigma (default): {'window': 7, 'sigma': 0.9, 'num_looks': 1, 'tk': 5}
                * 'sigma': float, (default: 0.9): Speckle noise standard deviation, applies for methods: ['lee_sigma']
                * 'num_looks': int, (default: 1): Number of looks of the SAR img, applies for methods: ['lee_sigma']
                * 'tk': int, (default: 5): Threshold of neighbouring pixels outside of the 98th percentile, applies for methods: ['lee_sigma']
                * cu : float, (default: 0.25): The noise variation coefficient, applies for methods: ['kuan','lee','lee_enhanced']
                * window : int, (default : 3): Denoising window. Increasing the window size may removes noise better but may also removes details of image in certain denoising methods.
        togray : bool, (default : False)
            Conversion to gray scale.
        verbose : int (default : 3)
            Print to screen. 0: None, 1: Error, 2: Warning, 3: Info, 4: Debug, 5: Trace.

        Returns
        -------
        dict()
            See 1dpeaks and 2dpeaks for more details.

        Examples
        --------
        >>> from findpeaks import findpeaks
        >>> X = [9,60,377,985,1153,672,501,1068,1110,574,135,23,3,47,252,812,1182,741,263,33]
        >>> fp = findpeaks(method='peakdetect', interpolate=10, lookahead=1)
        >>> results = fp.fit(X)
        >>> fp.plot()
        >>>
        >>> # 2D array example
        >>> from findpeaks import findpeaks
        >>> X = fp.import_example('2dpeaks')
        >>> results = fp.fit(X)
        >>> fp.plot()
        >>>
        >>> # Image example
        >>> from findpeaks import findpeaks
        >>> fp = findpeaks(method='topology', denoise='fastnl', params={'window': 30}, imsize=(300,300))
        >>> X = fp.import_example('2dpeaks_image')
        >>> results = fp.fit(X)
        >>> fp.plot()
        >>>
        >>> # Plot each seperately
        >>> fp.plot_preprocessing()
        >>> fp.plot_persistence()
        >>> fp.plot_mesh()

        References
        ----------
            * https://erdogant.github.io/findpeaks/
        """
        if window is not None: print('The input parameter "window" will be deprecated in future releases. Please use "params={"window": 5}" instead.')
        if cu is not None: print('The input parameter "cu" will be deprecated in future releases. Please use "params={"cu": 3}" instead.')

        # Store in object
        if isinstance(whitelist, str): whitelist=[whitelist]
        if lookahead is None: lookahead=1
        lookahead = np.maximum(1, lookahead)
        # if method is None: raise Exception('[findpeaks] >Specify the desired method="topology", "peakdetect", or "mask".')
        self.method = method
        self.whitelist = whitelist
        self.lookahead = lookahead
        self.interpolate = interpolate
        self.limit = limit
        self.imsize = imsize
        self.scale = scale
        self.togray = togray
        self.denoise = denoise
        self.figsize = figsize
        self.verbose = verbose

        # Store parameters for caerus
        defaults={}
        if method=='caerus':
            if len(params_caerus)>0:
                print('The input parameter "params_caerus" will be deprecated in future releases. Please use "params" instead.')
                params = params_caerus
            defaults = {'window': 50, 'minperc': 3, 'nlargest': 10, 'threshold': 0.25}
        elif method=='lee_sigma':
            defaults = {'window': 7, 'sigma': 0.9, 'num_looks': 1, 'tk': 5}
        defaults = {**{'window': 3}, **defaults}

        params = {**defaults, **params}
        self.window = params['window']
        self.cu = params.get('cu', 0.25)
        self.params = params

    def fit(self, X, x=None):
        """Detect peaks and valleys in a 1D vector or 2D-array (image).

        Description
        -----------
        * Fit the method on your data for the detection of peaks.
        * See 1dpeaks and 2dpeaks for more details about the input/output parameters.

        Parameters
        ----------
        X : array-like data.
            Input data.
        x : array-like data.
            Coordinates of the x-axis.

        Returns
        -------
        dict()
            * See 1dpeaks and 2dpeaks for more details.

        """
        # Check datatype
        if isinstance(X, list):
            X = np.array(X)
        if isinstance(X, type(pd.DataFrame())):
            X = X.values

        if len(X.shape)>1:
            # 2d-array (image)
            results = self.peaks2d(X, method=self.method)
        else:
            # 1d-array (vector)
            results = self.peaks1d(X, x=x, method=self.method)

        return results

    # Find peaks in 1D vector
    def peaks1d(self, X, x=None, method='peakdetect'):
        """Detect of peaks in 1D array.

        Description
        -----------
        This function only eats the input data. Use the .fit() function for more information regarding the input parameters:
            * method : method to be used for peak detection: 'topology' or 'peakdetect'.
            * lookahead : Looking ahead for peaks. For very small 1d arrays (such as up to 50 datapoints), use low numbers: 1 or 2.
            * interpolate : Interpolation factor. The higher the number, the less sharp the edges will be.
            * limit : Values > limit are set as regions of interest (ROI).
            * verbose : Print to screen. 0: None, 1: Error, 2: Warning, 3: Info, 4: Debug, 5: Trace.

        Parameters
        ----------
        X : array-like 1D vector.
            Input data.
        x : array-like 1D vector.
            Coordinates of the x-axis.

        Returns
        -------
        dict() : Results in "df" are based on the input-data, whereas "df_interp" are the interpolated results.
            * persistence : Scores when using topology method.
            * Xranked     : Similar to column "rank".
            * Xdetect     : Similar to the column "score".
            * df          : Is ranked in the same manner as the input data and provides information about the detected peaks and valleys.
        persistence : pd.DataFrame()
            * x, y    : coordinates
            * birth   : Birth level
            * death   : Death level
            * score   : persistence scores
        df : pd.DataFrame()
            * x, y    : Coordinates
            * labx    : The label of the peak area
            * rank    : The ranking number of the best performing peaks (1 is best)
            * score   : persistence score
            * valley  : Whether the point is marked as valley
            * peak    : Whether the point is marked as peak

        Examples
        --------
        >>> from findpeaks import findpeaks
        >>> X = [9,60,377,985,1153,672,501,1068,1110,574,135,23,3,47,252,812,1182,741,263,33]
        >>> fp = findpeaks(method='peakdetect', interpolate=10, lookahead=1)
        >>> results = fp.fit(X)
        >>> fp.plot()
        >>>
        >>> fp = findpeaks(method='topology')
        >>> results = fp.fit(X)
        >>> fp.plot()
        >>> fp.plot_persistence()

        """
        if method is None: method='peakdetect'
        self.method = method
        self.type = 'peaks1d'
        if self.verbose>=3: print('[findpeaks] >Finding peaks in 1d-vector using [%s] method..' %(self.method))
        # Make numpy array
        X = np.array(X)
        Xraw = X.copy()
        result = {}

        # Interpolation
        if self.interpolate is not None:
            X = interpolate.interpolate_line1d(X, n=self.interpolate, method=2, showfig=False, verbose=self.verbose)

        # Compute peaks based on method
        if method=='peakdetect':
            # Peakdetect method
            max_peaks, min_peaks = peakdetect(X, lookahead=self.lookahead)
            # Post processing for the peak-detect
            result['peakdetect'] = stats._post_processing(X, Xraw, min_peaks, max_peaks, self.interpolate, self.lookahead)
        elif method=='topology':
            # Compute persistence using toplogy method
            result = stats.topology(np.c_[X, X], limit=self.limit, verbose=self.verbose)
            # Post processing for the topology method
            result['topology'] = stats._post_processing(X, Xraw, result['valley'], result['peak'], self.interpolate, 1)
        elif method=='caerus':
            cs = caerus(**self.params)
            result = cs.fit(X, return_as_dict=True, verbose=self.verbose)
            # Post processing for the caerus method
            result['caerus'] = stats._post_processing(X, Xraw, np.c_[result['loc_start_best'], result['loc_start_best']], np.c_[result['loc_stop_best'], result['loc_stop_best']], self.interpolate, 1, labxRaw=result['df']['labx'].values)
            result['caerus']['model'] = cs
        else:
            if self.verbose>=2: print('[findpeaks] >WARNING: [method="%s"] is not supported in 1d-vector data. <return>' %(self.method))
            return None
        # Store
        self.results, self.args = self._store1d(X, Xraw, x, result)
        # Return
        return self.results

    # Store 1D vector
    def _store1d(self, X, Xraw, xs, result):
        # persist_score, res_peakd, results_topology
        # persist_score, results_peaksdetect, results_topology
        if xs is None: xs = np.arange(0, len(X))
        results = {}
        # Interpolated data
        dfint = pd.DataFrame()
        dfint['x'] = xs
        dfint['y'] = X
        # Store results for method
        if self.method=='peakdetect':
            # peakdetect
            dfint['labx'] = result['peakdetect']['labx_s']
            dfint['valley'] = False
            dfint['peak'] = False
            if result['peakdetect']['min_peaks_s'] is not None:
                dfint['valley'].iloc[result['peakdetect']['min_peaks_s'][:, 0].astype(int)] = True
            if result['peakdetect']['max_peaks_s'] is not None:
                dfint['peak'].iloc[result['peakdetect']['max_peaks_s'][:, 0].astype(int)] = True
        elif self.method=='topology':
            # Topology
            dfint['labx'] = result['topology']['labx_s']
            dfint['rank'] = result['Xranked']
            dfint['score'] = result['Xdetect']
            dfint['valley'] = False
            dfint['peak'] = False
            if result['topology']['min_peaks_s'] is not None:
                dfint['valley'].iloc[result['topology']['min_peaks_s'][:, 0].astype(int)] = True
            if result['topology']['max_peaks_s'] is not None:
                dfint['peak'].iloc[result['topology']['max_peaks_s'][:, 0].astype(int)] = True

            results['persistence'] = result['persistence']
            results['Xdetect'] = result['Xdetect']
            results['Xranked'] = result['Xranked']
            results['groups0'] = result['groups0']
        elif self.method=='caerus':
            # caerus
            dfint = result['df'].copy()
            dfint['y'] = result['X']
            dfint.drop(labels='X', inplace=True, axis=1)
            dfint['x'] = xs
            # dfint['labx'] = result['caerus']['labx_s']
            # dfint['valley'] = False
            # dfint['peak'] = False
            # if result['caerus']['min_peaks_s'] is not None:
            #     dfint['valley'].iloc[result['caerus']['min_peaks_s'][:, 0].astype(int)] = True
            # if result['caerus']['max_peaks_s'] is not None:
            #     dfint['peak'].iloc[result['caerus']['max_peaks_s'][:, 0].astype(int)] = True

        # As for the input data
        if self.interpolate is not None:
            df = pd.DataFrame()
            df['y'] = Xraw
            # Store results for method
            if self.method=='peakdetect':
                # peakdetect
                df['x'] = result['peakdetect']['xs']
                df['labx'] = result['peakdetect']['labx']
                df['valley'] = False
                df['peak'] = False
                if result['peakdetect']['min_peaks'] is not None:
                    df['valley'].iloc[result['peakdetect']['min_peaks'][:, 0].astype(int)] = True
                if result['peakdetect']['max_peaks'] is not None:
                    df['peak'].iloc[result['peakdetect']['max_peaks'][:, 0].astype(int)] = True
            elif self.method=='topology':
                # Topology
                df['x'] = result['topology']['xs']
                df['labx'] = result['topology']['labx']
                df['valley'] = False
                df['peak'] = False
                if result['topology']['min_peaks'] is not None:
                    df['valley'].iloc[result['topology']['min_peaks'][:, 0].astype(int)] = True
                if result['topology']['max_peaks'] is not None:
                    df['peak'].iloc[result['topology']['max_peaks'][:, 0].astype(int)] = True

                # Store the score and ranking
                df['rank'] = 0
                df['score'] = 0

                df['rank'].iloc[result['topology']['max_peaks'][:, 0].astype(int)] = dfint['rank'].iloc[result['topology']['max_peaks_s'][:, 0].astype(int)].values
                df['score'].iloc[result['topology']['max_peaks'][:, 0].astype(int)] = dfint['score'].iloc[result['topology']['max_peaks_s'][:, 0].astype(int)].values
                # df['rank'].loc[df['peak']] = dfint['rank'].loc[dfint['peak']].values
                # df['score'].loc[df['peak']] = dfint['score'].loc[dfint['peak']].values
            if self.method=='caerus':
                # caerus
                df['x'] = result['caerus']['xs']
                df['labx'] = result['df']['labx']
                df['valley'] = False
                df['peak'] = False
                if result['caerus']['min_peaks'] is not None:
                    df['valley'].iloc[result['caerus']['min_peaks'][:, 0].astype(int)] = True
                if result['caerus']['max_peaks'] is not None:
                    df['peak'].iloc[result['caerus']['max_peaks'][:, 0].astype(int)] = True

            # Store in results
            results['df'] = df
            results['df_interp'] = dfint
        else:
            results['df'] = dfint

        if self.method=='caerus':
            results['model'] = result['caerus']['model']
        # Arguments
        args = {}
        args['method'] = self.method
        args['params'] = self.params
        args['lookahead'] = self.lookahead
        args['interpolate'] = self.interpolate
        args['figsize'] = self.figsize
        args['type'] = self.type
        # Return
        return results, args

    # Find peaks in 2D-array
    def peaks2d(self, X, method='topology'):
        """Detect peaks and valleys in a 2D-array or image.

        Description
        -----------
        To handle 2d-arrays or images. Use the .fit() function for more information regarding the input parameters:
            * method : method to be used for peak detection: 'topology', or 'mask'
            * limit : Values > limit are set as regions of interest (ROI).
            * scale : Scaling data in range [0-255] by img*(255/max(img))
            * denoise : Remove noise using method:
                * None
                * 'fastnl'
                * 'bilateral'
                * 'lee'
                * 'lee_enhanced'
                * 'lee_sigma'
                * 'kuan'
                * 'frost'
                * 'median'
                * 'mean'
            * window : Denoising window.
            * cu : noise variation coefficient
            * togray : Conversion to gray scale.
            * imsize : resize image
            * verbose : Print to screen. 0: None, 1: Error, 2: Warning, 3: Info, 4: Debug, 5: Trace.

        Parameters
        ----------
        X : array-like 1D vector.
            Input data.

        Returns
        -------
        dict()
            * Xraw    : The RAW input data
            * Xproc   : The pre-processed data
            * Xdetect : The detected peaks with the persistence scores (same shape as the input data)
            * XRanked : The detected peaks but based on the strenght (same shape as the input data)
            * persistence : pd.DataFrame()
                * x, y    : coordinates
                * birth   : Birth level
                * death   : Death level
                * score   : persistence scores

        Examples
        --------
        >>> # 2D array example
        >>> from findpeaks import findpeaks
        >>> X = fp.import_example('2dpeaks')
        >>> results = fp.fit(X)
        >>> fp.plot()
        >>>
        >>> # Image example
        >>> from findpeaks import findpeaks
        >>> X = fp.import_example('2dpeaks_image')
        >>> fp = findpeaks(denoise='fastnl', params={'window': 30}, imsize=(300,300))
        >>> results = fp.fit(X)
        >>> fp.plot()
        >>>
        >>> # Plot each seperately
        >>> fp.plot_preprocessing()
        >>> fp.plot_persistence()
        >>> fp.plot_mesh()

        """
        if method is None: method='topology'
        self.method = method
        self.type = 'peaks2d'
        if self.verbose>=3: print('[findpeaks] >Finding peaks in 2d-array using %s method..' %(self.method))
        if (not self.togray) and (len(X.shape)==3) and (self.method=='topology'): raise Exception('[findpeaks] >Error: Topology method requires 2d-array. Your input is 3d. Hint: set togray=True.')

        # Preprocessing the image
        Xproc = self.preprocessing(X, showfig=False)
        # Compute peaks based on method
        if method=='topology':
            # Compute persistence based on topology method
            result = stats.topology2d(Xproc, limit=self.limit, whitelist=self.whitelist, verbose=self.verbose)
            # result = stats.topology(Xproc, limit=self.limit, verbose=self.verbose)
        elif method=='mask':
            # Compute peaks using local maximum filter.
            result = stats.mask(Xproc, limit=self.limit, verbose=self.verbose)
        else:
            if self.verbose>=2: print('[findpeaks] >WARNING: [method="%s"] is not supported in 2d-array (image) data. <return>' %(self.method))
            return None

        # Store
        self.results, self.args = self._store2d(X, Xproc, result)
        # Return
        if self.verbose>=3: print('[findpeaks] >Fin.')
        return self.results

    # Store 2D-array
    def _store2d(self, X, Xproc, result):
        # Store results
        results = {}
        results['Xraw'] = X
        results['Xproc'] = Xproc

        # Store method specific results
        if self.method=='topology':
            # results['topology'] = result
            results['Xdetect'] = result['Xdetect']
            results['Xranked'] = result['Xranked']
            results['persistence'] = result['persistence']
            # results['peak'] = result['peak'] # These values are incorrect when using 2d
            # results['valley'] = result['valley'] # These values are incorrect when using 2d
            results['groups0'] = result['groups0']
        if self.method=='mask':
            results['Xdetect'] = result['Xdetect']
            results['Xranked'] = result['Xranked']

        # Store arguments
        args = {}
        args['limit'] = self.limit
        args['scale'] = self.scale
        args['denoise'] = self.denoise
        args['togray'] = self.togray
        args['imsize'] = self.imsize
        args['figsize'] = self.figsize
        args['type'] = self.type
        # Return
        return results, args

    # Pre-processing
    def preprocessing(self, X, showfig=False):
        """Preprocessing steps of the 2D array (image).

        The pre-processing has 4 (optional) steps.
            1. Resizing (to reduce computation time).
            2. Scaling color pixels between [0-255]
            3. Conversion to gray-scale. This is required for some analysis.
            4. Denoising of the image.

        Parameters
        ----------
        X : numpy-array
            Input data or image.
        showfig : bool
            Show the preocessing steps in figures. The default is None.

        Returns
        -------
        X : numpy-array
            Processed image.

        """
        if showfig:
            # Number of axis to create:
            nplots = 1 + (self.imsize is not None) + self.scale + self.togray + (self.denoise is not None)
            fig, ax = plt.subplots(1, nplots, figsize=self.figsize)
            iax = 0

            # Plot RAW input image
            ax[iax].imshow(X, cmap=('gray_r' if self.togray else None))
            ax[iax].grid(False)
            ax[iax].set_title('Input\nRange: [%.3g,%.3g]' %(X.min(), X.max()))
            iax = iax + 1
            # plt.show()

        # Resize
        if self.imsize:
            X = stats.resize(X, size=self.imsize, verbose=self.verbose)
            if showfig:
                # plt.figure(figsize=self.figsize)
                ax[iax].imshow(X, cmap=('gray_r' if self.togray else None))
                ax[iax].grid(False)
                ax[iax].set_title('Resize\n(%s,%s)' %(self.imsize))
                iax = iax + 1
        # Scaling color range between [0,255]
        if self.scale:
            X = stats.scale(X, verbose=self.verbose)
            if showfig:
                # plt.figure(figsize=self.figsize)
                ax[iax].imshow(X, cmap=('gray_r' if self.togray else None))
                ax[iax].grid(False)
                ax[iax].set_title('Scale\nRange: [%.3g %.3g]' %(X.min(), X.max()))
                iax = iax + 1
        # Convert to gray image
        if self.togray:
            X = stats.togray(X, verbose=self.verbose)
            if showfig:
                # plt.figure(figsize=self.figsize)
                ax[iax].imshow(X, cmap=('gray_r' if self.togray else None))
                ax[iax].grid(False)
                ax[iax].set_title('Color conversion\nGray')
                iax = iax + 1
        # Denoising
        if self.denoise is not None:
            X = stats.denoise(X, method=self.denoise, window=self.window, cu=self.cu, verbose=self.verbose)
            if showfig:
                # plt.figure(figsize=self.figsize)
                ax[iax].imshow(X, cmap=('gray_r' if self.togray else None))
                ax[iax].grid(False)
                ax[iax].set_title('Denoise\n' + self.method)
                iax = iax + 1
        # Return
        return X

    # Pre-processing
    def imread(self, path, verbose=3):
        """Read file from disk or url.

        Parameters
        ----------
        path : String
            filepath or Url.

        Returns
        -------
        X : Numpy array

        """
        cv2 = stats._import_cv2()
        if is_url(path):
            if verbose>=3: print('[findpeaks] >Downloading from github source: [%s]' %(path))
            response = requests.get(path)
            img_array = np.asarray(bytearray(response.content), dtype=np.uint8)
            X = cv2.imdecode(img_array, cv2.IMREAD_COLOR)
        elif os.path.isfile(path):
            if verbose>=3: print('[findpeaks] >Import [%s]' %(path))
            X = cv2.imread(path)
        # Return
        return X

    # %% Plotting
    def plot(self, limit=None, legend=True, figsize=None, cmap=None, text=True, xlabel='x-axis', ylabel='y-axis'):
        """Plot results.

        Parameters
        ----------
        legend : bool, (default: True)
            Show the legend.
        figsize : (int, int), optional, default: (15, 8)
            (width, height) in inches.
        cmap : object (default : None)
            Colormap. The default is derived wether image is convert to grey or not. Other options are: plt.cm.hot_r.
        text : Bool (default : True)
            Include text to the 2d-image that shows the peaks (p-number) and valleys (v-number)

        Returns
        -------
        fig_axis : tuple containing (fig, ax)

        """
        if not hasattr(self, 'results'):
            if self.verbose>=2: print('[findpeaks] >WARNING: Nothing to plot. <return>')
            return None

        figsize = figsize if figsize is not None else self.args['figsize']

        if self.args['type']=='peaks1d':
            fig_axis = self.plot1d(legend=legend, figsize=figsize, xlabel=xlabel, ylabel=ylabel)
        elif self.args['type']=='peaks2d':
            # fig_axis = self.plot2d(figsize=figsize)
            fig_axis = self.plot_mask(figsize=figsize, cmap=cmap, text=text, limit=limit)
        else:
            if self.verbose>=2: print('[findpeaks] >WARNING: Nothing to plot for %s' %(self.args['type']))
            return None

        # Return
        return fig_axis

    def plot1d(self, legend=True, figsize=None, xlabel='x-axis', ylabel='y-axis'):
        """Plot the 1D results.

        Parameters
        ----------
        legend : bool, (default: True)
            Show the legend.
        figsize : (int, int), (default: None)
            (width, height) in inches.

        Returns
        -------
        fig_axis : tuple containing axis for each figure.

        """
        if not self.args['type']=='peaks1d':
            if self.verbose>=3: print('[findpeaks] >Requires results of 1D data <return>.')
            return None

        figsize = figsize if figsize is not None else self.args['figsize']
        ax1, ax2 = None, None
        title = self.method

        if self.method=='caerus':
            if self.results.get('model', None) is not None:
                ax = self.results['model'].plot(figsize=self.figsize)
                csplots._plot_graph(self.results['model'].results, figsize=self.figsize, xlabel=xlabel, ylabel=ylabel)
                # Return axis
                return ax
        else:
            # Make plot
            min_peaks, max_peaks = np.array([]), np.array([])
            df = self.results['df']
            if np.any('valley' in self.whitelist):
                min_peaks = df['x'].loc[df['valley']].values
            if np.any('peak' in self.whitelist):
                max_peaks = df['x'].loc[df['peak']].values
            ax1 = _plot_original(df['y'].values, df['x'].values, df['labx'].values, min_peaks.astype(int), max_peaks.astype(int), title=title, figsize=figsize, legend=legend, xlabel=xlabel, ylabel=ylabel)

            # Make interpolated plot
            if self.interpolate is not None:
                min_peaks, max_peaks = np.array([]), np.array([])
                df_interp = self.results['df_interp']
                if np.any('valley' in self.whitelist):
                    min_peaks = df_interp['x'].loc[df_interp['valley']].values
                if np.any('peak' in self.whitelist):
                    max_peaks = df_interp['x'].loc[df_interp['peak']].values
                ax2 = _plot_original(df_interp['y'].values, df_interp['x'].values, df_interp['labx'].values, min_peaks.astype(int), max_peaks.astype(int), title=title + ' (interpolated)', figsize=figsize, legend=legend, xlabel=xlabel, ylabel=ylabel)
            # Return axis
            return (ax2, ax1)

    def plot2d(self, figsize=None, limit=None):
        """Plot the 2d results.

        Parameters
        ----------
        figsize : (int, int), (default: None)
            (width, height) in inches.

        Returns
        -------
        fig_axis : tuple containing axis for each figure.

        """
        if not self.args['type']=='peaks2d':
            if self.verbose>=3: print('[findpeaks] >Requires results of 2D data <return>.')
            return None
        ax_method, ax_mesh = None, None
        figsize = figsize if figsize is not None else self.args['figsize']
        # Plot preprocessing steps
        self.plot_preprocessing()

        # Setup figure
        if self.method=='mask':
            ax_method = self.plot_mask(figsize=figsize, limit=limit)
        if self.method=='topology':
            # Plot topology/persistence
            ax_method = self.plot_persistence(figsize=figsize)

        # Plot mesh
        ax_mesh = self.plot_mesh(figsize=figsize)

        # Return axis
        return (ax_method, ax_mesh)

    def plot_preprocessing(self):
        """Plot the pre-processing steps.

        Returns
        -------
        None.

        """
        if (not hasattr(self, 'results')) or (self.type=='peaks1d'):
            if self.verbose>=2: print('[findpeaks] >WARNING: Nothing to plot. Hint: run fit(X), where X is the (image) data. <return>')
            return None

        _ = self.preprocessing(X=self.results['Xraw'], showfig=True)

    def plot_mask(self, limit=None, figsize=None, cmap=None, text=True):
        """Plot the masking.

        Parameters
        ----------
        limit : float, (default : None)
            Values > limit are set as regions of interest (ROI).
        figsize : (int, int), (default: None)
            (width, height) in inches.
        cmap : object (default : None)
            Colormap. The default is derived wether image is convert to grey or not. Other options are: plt.cm.hot_r.

        Returns
        -------
        fig_axis : tuple containing axis for each figure.

        """
        if (self.type=='peaks1d'):
            if self.verbose>=2: print('[findpeaks] >WARNING: Nothing to plot. Hint: run fit(X), where X is the 2d-array (image). <return>')
            return None

        if limit is None: limit = self.limit
        # Only show above the limit
        Xdetect = self.results['Xdetect'].copy()
        if limit is not None:
            Xdetect[np.abs(Xdetect)<limit]=0
        if cmap is None:
            cmap = 'gray' if self.args['togray'] else None
            cmap = cmap + '_r'
        # Get the index for the detected peaks/valleys
        idx_peaks = np.where(Xdetect>0)
        idx_valleys = np.where(Xdetect<0)

        # Setup figure
        figsize = figsize if figsize is not None else self.args['figsize']
        fig, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=figsize)

        # Plot input image
        ax1.imshow(self.results['Xraw'], cmap, interpolation="nearest")
        ax1.set_title('Input')
        ax1.grid(False)

        # For vizualisation purposes, plot all absolute numbers
        Xproc = self.results['Xproc'].copy()
        Xproc[idx_peaks]=0
        Xproc[idx_valleys]=1
        ax2.imshow(Xproc, cmap, interpolation="nearest")
        ax2.set_title('Processed image')
        ax2.grid(False)

        # Masking
        ax3.imshow(np.abs(Xdetect), 'gray_r', interpolation="nearest")
        ax3.set_title(self.method + ' (' + str(len(np.where(Xdetect>0)[0])) + ' peaks and ' + str(len(np.where(Xdetect<0)[0])) + ' valleys)')
        ax3.grid(False)

        if text:
            for idx in tqdm(zip(idx_peaks[0], idx_peaks[1]), disable=disable_tqdm(self.verbose)):
                ax2.text(idx[1], idx[0], 'p' + self.results['Xranked'][idx].astype(str))
                ax3.text(idx[1], idx[0], 'p' + self.results['Xranked'][idx].astype(str))

            for idx in tqdm(zip(idx_valleys[0], idx_valleys[1]), disable=disable_tqdm(self.verbose)):
                ax2.text(idx[1], idx[0], 'v' + self.results['Xranked'][idx].astype(str))
                ax3.text(idx[1], idx[0], 'v' + self.results['Xranked'][idx].astype(str))

        # Show plot
        plt.show()
        # Return
        return (ax1, ax2, ax3)

    def plot_mesh(self,
                  wireframe=True,
                  surface=True,
                  rstride=2,
                  cstride=2,
                  cmap=plt.cm.hot_r,
                  view=None,
                  xlim=None,
                  ylim=None,
                  zlim=None,
                  title='',
                  figsize=None,
                  savepath=None):
        """Plot the 3d-mesh.

        Parameters
        ----------
        wireframe : bool, (default is True)
            Plot the wireframe
        surface : bool, (default is True)
            Plot the surface
        rstride : int, (default is 2)
            Array row stride (step size).
        cstride : int, (default is 2)
            Array column stride (step size).
        figsize : (int, int), optional, default: (15, 8)
            (width, height) in inches.
        view : tuple, (default : None)
            * Rotate the mesh plot.
            * (0, 0) : y vs z
            * (0, 90) : x vs z
            * (90, 0) : y vs x
            * (90, 90) : x vs y
        cmap : object
            Colormap. The default is plt.cm.hot_r.
        xlim : tuple(int, int), (default: None)
            x-limit in the axis.
            None: No limit.
            [1, 5]: Limit between the range 1 and 5.
            [1, None]: Limit between range 1 and unlimited.
            [None, 5]: Limit between range unlimited and 5.
        ylim : tuple(int, int), (default: None)
            y-limit in the axis.
            None: No limit.
            [1, 5]: Limit between the range 1 and 5.
            [1, None]: Limit between range 1 and unlimited.
            [None, 5]: Limit between range unlimited and 5.
        zlim : tuple(int, int), (default: None)
            z-limit in the axis.
            None: No limit.
            [1, 5]: Limit between the range 1 and 5.
            [1, None]: Limit between range 1 and unlimited.
            [None, 5]: Limit between range unlimited and 5.
        figsize : (int, int), (default: None)
            (width, height) in inches.
        savepath : bool (default : None)
            Path with filename to save the figure, eg: './tmp/my_image.png'
        verbose : int (default : 3)
            Print to screen. 0: None, 1: Error, 2: Warning, 3: Info, 4: Debug, 5: Trace.
        
        Example
        -------
        >>> # Import library
        >>> from findpeaks import findpeaks
        >>> #
        >>> # Initialize
        >>> fp = findpeaks(method='topology', scale=False, denoise=None, togray=False, imsize=False, params={'window': 15})
        >>> #
        >>> # Load example data set
        >>> X = fp.import_example('2dpeaks')
        >>> #
        >>> # Fit model
        >>> fp.fit(X)
        >>> #
        >>> # Create mesh plot
        >>> fp.plot_mesh()
        >>> # Create mesh plot with limit on x-axis and y-axis
        >>> fp.plot_mesh(xlim=[10, 30], ylim=[4, 10], zlim=[None, 8])

        Returns
        -------
        fig_axis : tuple containing axis for each figure.

        """
        if not hasattr(self, 'results'):
            if self.verbose>=2: print('[findpeaks] >WARNING: Nothing to plot. Hint: run the fit() function. <return>')
            return None
        if self.results.get('Xproc', None) is None:
            if self.verbose>=3: print('[findpeaks] >These analysis do not support mesh plotting. This may be caused because your are analysing 1D.')
            return None

        figsize = figsize if figsize is not None else self.args['figsize']
        if self.verbose>=3: print('[findpeaks] >Plotting 3d-mesh..')
        ax1, ax2 = None, None
        if savepath is not None:
            savepath = str.replace(savepath, ',', '_')
            savepath = str.replace(savepath, '=', '_')

        # Compute meshgrid
        Z = self.results['Xproc'].copy()
        X, Y = np.mgrid[0:Z.shape[0], 0:Z.shape[1]]
        # To limit on the X and Y axis, we need to create a trick by setting all values to nan in the Z-axis that should be limited.
        if xlim is not None:
            if xlim[0] is not None: Z[X<xlim[0]]=np.nan
            if xlim[1] is not None: Z[X>xlim[1]]=np.nan
        if ylim is not None:
            if ylim[0] is not None: Z[Y<ylim[0]]=np.nan
            if ylim[1] is not None: Z[Y>ylim[1]]=np.nan
        if zlim is not None:
            if zlim[0] is not None: Z[Z<zlim[0]]=np.nan
            if zlim[1] is not None: Z[Z>zlim[1]]=np.nan

        # Plot the figure
        if wireframe:
            fig = plt.figure(figsize=figsize)
            ax1 = fig.add_subplot(projection='3d')
            ax1 = fig.gca()
            ax1.plot_wireframe(X, Y, Z, rstride=rstride, cstride=cstride, linewidth=0.8)
            ax1.set_xlabel('x-axis')
            ax1.set_ylabel('y-axis')
            ax1.set_zlabel('z-axis')
            if view is not None:
                ax1.view_init(view[0], view[1])
                # ax1.view_init(50, -10) # x vs y
            ax1.set_title(title)
            if xlim is not None: ax1.set_xlim3d(xlim[0], xlim[1])
            if ylim is not None: ax1.set_ylim3d(ylim[0], ylim[1])
            if zlim is not None: ax1.set_zlim3d(zlim[0], zlim[1])

            plt.show()
            if savepath is not None:
                if self.verbose>=3: print('[findpeaks] >Saving wireframe to disk..')
                fig.savefig(savepath)

        if surface:
            # Plot the figure
            fig = plt.figure(figsize=figsize)
            ax2 = fig.add_subplot(projection='3d')
            ax2 = fig.gca()
            ax2.plot_surface(X, Y, Z, rstride=rstride, cstride=cstride, cmap=cmap, linewidth=0, shade=True, antialiased=False)
            if view is not None:
                ax2.view_init(view[0], view[1])
            ax2.set_xlabel('x-axis')
            ax2.set_ylabel('y-axis')
            ax2.set_zlabel('z-axis')
            ax2.set_title(title)
            if xlim is not None: ax2.set_xlim3d(xlim[0], xlim[1])
            if ylim is not None: ax2.set_ylim3d(ylim[0], ylim[1])
            if zlim is not None: ax2.set_zlim3d(zlim[0], zlim[1])
            plt.show()
            if savepath is not None:
                if self.verbose>=3: print('[findpeaks] >Saving surface to disk..')
                fig.savefig(savepath)

        # Plot with contours
        # fig = plt.figure(figsize=figsize)
        # ax3 = fig.gca(projection='3d')
        # X, Y, Z = results['xx'], results['yy'], results['Xproc']
        # ax3.plot_surface(results['xx'], results['yy'], results['Xproc'], rstride=rstride, cstride=cstride, cmap=plt.cm.coolwarm, linewidth=0, shade=True, alpha=0.3)
        # cset = ax3.contour(X, Y, Z, zdir='z', offset=-100, cmap=plt.cm.coolwarm)
        # cset = ax3.contour(X, Y, Z, zdir='x', offset=-40, cmap=plt.cm.coolwarm)
        # cset = ax3.contour(X, Y, Z, zdir='y', offset=40, cmap=plt.cm.coolwarm)
        # plt.show()
        return ax1, ax2

    def plot_persistence(self, figsize=(20, 8), fontsize_ax1=14, fontsize_ax2=14, xlabel='x-axis', ylabel='y-axis', verbose=None):
        """Plot the homology-peristence.

        Parameters
        ----------
        figsize : (int, int), (default: None)
            (width, height) in inches.
        fontsize_ax1 : int, (default: 14)
            Font size for the labels in the left figure. Choose None for no text-labels.
        fontsize_ax2 : int, (default: 14)
            Font size for the labels in the right figure. Choose None for no text-labels.
        verbose : int (default : 3)
            Print to screen. 0: None, 1: Error, 2: Warning, 3: Info, 4: Debug, 5: Trace.

        Returns
        -------
        ax1 : object
            Figure axis 1.
        ax2 : object
            Figure axis 2.

        """
        if verbose is None: verbose=self.verbose
        if (self.method!='topology') or (not hasattr(self, 'results')):
            if verbose>=3: print('[findpeaks] >WARNING: Nothing to plot. Hint: run the .fit(method="topology") function. <return>')
            return None

        # Setup figure
        figsize = figsize if figsize is not None else self.args['figsize']
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=figsize)

        # Create the persistence ax2
        ax2 = self._plot_persistence_ax2(fontsize_ax2, ax2, verbose)
        # Create the persistence ax1
        ax1, ax2 = self._plot_persistence_ax1(fontsize_ax1, ax1, ax2, figsize, xlabel, ylabel, verbose)
        # Plot
        plt.show()
        # Return
        return ax1, ax2

    def _plot_persistence_ax1(self, fontsize, ax1, ax2, figsize, xlabel, ylabel, verbose):
        if self.args['type']=='peaks1d':
            # Attach the ranking-labels
            if fontsize is not None:
                y = self.results['df']['y'].values
                x = self.results['df']['x'].values
                idx = np.where(self.results['df']['rank']>0)[0]
                for i in tqdm(idx, disable=disable_tqdm(verbose)):
                    ax1.text(x[i], (y[i] + y[i] * 0.01), str(self.results['df']['rank'].iloc[i]), color='b', fontsize=fontsize)

            # minpers = 0
            min_peaks, max_peaks = np.array([]), np.array([])
            if np.any('valley' in self.whitelist):
                min_peaks = self.results['df']['x'].loc[self.results['df']['valley']].values
            if np.any('peak' in self.whitelist):
                max_peaks = self.results['df']['x'].loc[self.results['df']['peak']].values
            # Make the plot
            ax1 = _plot_original(self.results['df']['y'].values, self.results['df']['x'].values, self.results['df']['labx'].values, min_peaks.astype(int), max_peaks.astype(int), title='Persistence', figsize=figsize, legend=True, ax=ax1, xlabel=xlabel, ylabel=ylabel)
            # Set limits
            X = np.c_[self.results['df']['x'].values, self.results['df']['y'].values]
            ax1.set_xlim((np.min(X), np.max(X)))
            ax1.set_ylim((np.min(X), np.max(X)))
        else:
            # X = self.results['Xproc']
            # Make the figure
            Xdetect = np.zeros_like(self.results['Xproc']).astype(int)
            # fig, ax1 = plt.subplots()
            # minpers = 1
            # Plot the detected loci
            if verbose>=3: print('[findpeaks] >Plotting loci of birth..')
            ax1.set_title("Loci of births")
            for i, homclass in tqdm(enumerate(self.results['groups0']), disable=disable_tqdm(verbose)):
                p_birth, bl, pers, p_death = homclass
                if (self.limit is None):
                    y, x = p_birth
                    Xdetect[y, x] = i + 1
                    ax1.plot([x], [y], '.', c='b')
                    ax1.text(x, y + 0.25, str(i), color='b', fontsize=fontsize)
                elif pers > self.limit:
                    y, x = p_birth
                    Xdetect[y, x] = i + 1
                    ax1.plot([x], [y], '.', c='b')
                    ax1.text(x, y + 0.25, str(i), color='b', fontsize=fontsize)

            ax1.set_xlim((0, self.results['Xproc'].shape[1]))
            ax1.set_ylim((0, self.results['Xproc'].shape[0]))
            ax1.invert_yaxis()
            plt.gca().invert_yaxis()
            ax1.grid(True)
            # Set the axis to 255-255 in ax2 because it is an image.
            ax2.plot([0, 255], [0, 255], '-', c='grey')
        return ax1, ax2

    def _plot_persistence_ax2(self, fontsize, ax2, verbose):
        x = self.results['persistence']['birth_level'].values
        y = self.results['persistence']['death_level'].values
        ax2.plot(x, y, '.', c='b')
        if fontsize is not None:
            for i in tqdm(range(0, len(x)), disable=disable_tqdm(verbose)):
                ax2.text(x[i], (y[i] + y[i] * 0.01), str(i + 1), color='b', fontsize=fontsize)

        X = np.c_[x, y]
        ax2.plot([np.min(X), np.max(X)], [np.min(X), np.max(X)], '-', c='grey')
        ax2.set_xlabel("Birth level")
        ax2.set_ylabel("Death level")
        ax2.set_xlim((np.min(X), np.max(X)))
        ax2.set_ylim((np.min(X), np.max(X)))
        ax2.grid(True)
        return ax2

    def import_example(self, data='2dpeaks', url=None, sep=';', datadir=None):
        """Import example dataset from github source.

        Description
        -----------
        Import one of the few datasets from github source or specify your own download url link.

        Parameters
        ----------
        data : str
            Name of datasets: "1dpeaks", "2dpeaks", "2dpeaks_image", 'btc', 'facebook'
        url : str
            url link to to dataset.
        Verbose : int (default : 3)
            Print to screen. 0: None, 1: Error, 2: Warning, 3: Info, 4: Debug, 5: Trace.
        datadir : path-like
            Directory to store downloaded datasets in. Defaults to data sub-directory
            of findpeaks install location.

        Returns
        -------
        pd.DataFrame()
            Dataset containing mixed features.

        """
        X = import_example(data=data, url=url, sep=sep, verbose=self.verbose, datadir=datadir)
        return X


# %%
def _plot_original(X, xs, labx, min_peaks, max_peaks, title=None, legend=True, ax=None, figsize=(15, 8), xlabel=None, ylabel=None):
    uilabx = np.unique(labx)
    uilabx = uilabx[~np.isnan(uilabx)]

    if ax is None: fig, ax = plt.subplots(figsize=figsize)
    ax.plot(xs, X, 'k')
    if np.any(max_peaks):
        ax.plot(max_peaks, X[max_peaks], "x", label='Peak')
    if np.any(min_peaks):
        ax.plot(min_peaks, X[min_peaks], "o", label='Valley')

    # Color each detected label
    s=np.arange(0, len(X))
    for i in uilabx:
        idx=(labx==i)
        ax.plot(s[idx], X[idx])
        # plt.plot(s[idx], X[idx], label='peak' + str(i))

    if legend: ax.legend(loc=0)
    ax.set_title(title)
    if xlabel is not None: ax.set_xlabel(xlabel)
    if ylabel is not None: ax.set_ylabel(ylabel)
    ax.grid(True)
    plt.show()
    return ax


# %% Import example dataset from github.
def import_example(data='2dpeaks', url=None, sep=';', verbose=3, datadir=None):
    """Import example dataset from github source.

    Description
    -----------
    Import one of the few datasets from github source or specify your own download url link.

    Parameters
    ----------
    data : str
        Name of datasets: "2dpeaks" or "2dpeaks_image"
    url : str
        url link to to dataset.
    Verbose : int (default : 3)
        Print to screen. 0: None, 1: Error, 2: Warning, 3: Info, 4: Debug, 5: Trace.
    datadir : path-like
        Directory to store downloaded datasets in. Defaults to data sub-directory
        of findpeaks install location.

    Returns
    -------
    pd.DataFrame()
        Dataset containing mixed features.

    """
    if url is not None:
        fn = os.path.basename(urlparse(url).path).strip()
        if not fn:
            if verbose>=3: print('[findpeaks] >Could not determine filename to download <return>.')
            return None
        data, _ = os.path.splitext(fn)
    elif data=='2dpeaks_image':
        url='https://erdogant.github.io/datasets/' + data + '.png'
        fn = "2dpeaks_image.png"
    elif data=='2dpeaks':
        url='https://erdogant.github.io/datasets/' + data + '.zip'
        fn = "2dpeaks.zip"
    elif data=='1dpeaks':
        # x = [0,   13,  22,  30,  35,  38,   42,   51,   57,   67,  73,   75,  89,   126,  141,  150,  200 ]
        y = [1.5, 0.8, 1.2, 0.2, 0.4, 0.39, 0.42, 0.22, 0.23, 0.1, 0.11, 0.1, 0.14, 0.09, 0.04, 0.02, 0.01]
        # X = np.c_[x, y]
        return y
    elif (data=='btc') or (data=='facebook'):
        from caerus import caerus
        cs = caerus()
        X = cs.download_example(name=data, verbose=verbose)
        return X
    else:
        if verbose>=2: print('[findpeaks] >WARNING: Nothing to download. <return>')
        return None

    if datadir is None: datadir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
    PATH_TO_DATA = os.path.join(datadir, fn)
    if not os.path.isdir(datadir):
        os.makedirs(datadir, exist_ok=True)

    # Check file exists.
    if not os.path.isfile(PATH_TO_DATA):
        if verbose>=3: print('[findpeaks] >Downloading from github source: [%s]' %(url))
        r = requests.get(url, stream=True)
        with open(PATH_TO_DATA, "wb") as fd:
            for chunk in r.iter_content(chunk_size=1024):
                fd.write(chunk)

    # Import local dataset
    if verbose>=3: print('[findpeaks] >Import [%s]' %(PATH_TO_DATA))
    if data=='2dpeaks_image':
        cv2 = stats._import_cv2()
        X = cv2.imread(PATH_TO_DATA)
    else:
        X = pd.read_csv(PATH_TO_DATA, sep=sep).values
    # Return
    return X


# Check url
def is_url(url):
    try:
        _ = urlparse(url)
        return True
    except ValueError:
        return False

# %%
# def disable_tqdm(verbose):
    # """Set the verbosity messages."""
    # return  (True if ((verbose==0 or verbose is None) or verbose>3) else False)
